/****************************************************************************
 **                                                                        **
 **    File:           cparse.c                                            **
 **    Author:         Roger Preston (modifications only)                  **
 **    Created/Copied: Mon Dec 19 11:21:27 EST 1988                        **
 **    Contents :      Command parser from the Dead Box Dumper's command   **
 **                    processor; implements a standalone console command  **
 **                    processor                                           **
 **                                                                        **
 **    Copyright (c) 1988 BBN Communications Corp.                         **
 **    All rights reserved.                                                **
 **                                                                        **
 ****************************************************************************/
#include "tmp4.h"
#include "tmp3.h"
#include "tmp1.h"

extern char *gracie_version;
extern char *gracie_date;

extern int cpu_getc();
extern int cpu_putc();
extern int cpu_kbhit();

extern int g_tcs_addr;
extern struct dump_cmd_definition dcmd_def[];
extern struct cpid pid_array[];


/*****************************************************************************
 *                                                                           *
 * This file contains the parser and a few basic commands for a              *
 * stand-alone diagnostic monitor that runs on the Butterfly 2 TCS           *
 * Master.   Commands are listed in a structure called dcmd_def,             *
 * which gives the name for each command and the associated subroutine       *
 * to execute.  The commands in this file are used mainly for debugging      *
 * the debugger, as it were.                                                 *
 *                                                                           *
 *  Modified from the BBN BPAD by g fedorkow, june/86                        *
 *  and again and again, most recently by Bob Clements, 1988.                *
 *                                                                           *
 *****************************************************************************/


/* Top level routine called from main */

command_scanner()
{
    application_init();        /* Init. variables specific to application */
    console_cmd_init();        /* Init. console */

    while(TRUE) {
	if (cpu_kbhit())
    	    prepare_cmd();
    }
}



/*****************************************************************************
 *                                                                           *
 * Dumper Command Processor - This routine is called for each character      *
 * input from the console teletype.  Upon receipt of a carriage return       *
 * it tries to parse the command.  TRUE is returned by this                  *
 * routine to indicate that the command can be executed.                     *
 *                                                                           *
 *****************************************************************************/
struct input_line input;

prepare_cmd()
{
  char rchar;
  
  rchar = cpu_getc();
  if (get_cmd(rchar, &input) == TRUE)
  {
    do_command(&input);
    do_dump_prompt();
  }
}




do_command(ilp)
register struct input_line *ilp;
{
  long pass_mask;
  
  if (ilp->cm_type == UNDEFINED) 
    printf("unknown command: '%s'\n", ilp->buf);
  else if (ilp->cm_type != NULL_DUMP_CMD) 
  {
    pass_mask = dcmd_def[ilp->cm_type].pass_msg_count;
    
    /*
     * We're going to run the command at least once.  If the 'loop'
     * flag is set, we'll re-run it until either it returns
     * false, or the user types any character on the tty
     * The loop flag has three states, corresponding to running
     * the command a single time, then stopping, a state for
     * the first pass through of a looping command, and a third
     * state for the 'nth' pass through.
     * If the command is marked 'non-stop', it will be re-run
     * regardless of the return code.
     */
    
    ilp->pass_count = 0;
    do {
      if (((*dcmd_def[ilp->cm_type].cmd_func)(ilp) == FALSE) &&
	  (ilp->non_stop == FALSE))
	ilp->loop = DONT_LOOP;
      
      if (ilp->loop != DONT_LOOP) 
      {
	if (cpu_getc() != UNDEFINED)
	  ilp->loop = DONT_LOOP;
      }
      /* update the state for the second and subsequent looping
       * passes
       */
      if (ilp->loop == FIRST_TIME)
	ilp->loop = MANY_TIMES;
      ilp->pass_count++;
      if ((ilp->pass_count & pass_mask) == 0)
	printf("%s pass %d\n", ilp->arg[0], ilp->pass_count);
    } while (ilp->loop != DONT_LOOP);
  }
}



/*****************************************************************************
 * get_cmd -- check each character as it's typed; return TRUE if RETURN      *
 *            has been entered and command is ready for parsing.             *
 *****************************************************************************/
get_cmd(rchar, ilp)
char rchar;
register struct input_line *ilp;
{
  register int i;
  
  rchar &= 0x7f;
  
  switch (rchar)
  {
    case 'U' - '\100':      /*line kill */
      printf(" #\n");
      *ilp->buf = '\0';
      ilp->index = 0;
      return(FALSE);
    
    case 'H' - '\100':       /*backspace */
    case '\177':
      if (ilp->index == 0)
        break;
      ilp->index--;
      printf("\010 \010", ilp->buf[ilp->index]);
      return(FALSE);
    
    case 'R' - '\100':       /* repeat command */
        /* scopy(ilp->last_cmd, ilp->buf);  old stuff */
      for (i = 0; i < ilp->argc; i++)
        printf("%s ", ilp->arg[i]);
      printf("\n");
      ilp->index = 0;
      return(TRUE);
    
    case '\r':
    case '\n':
      printf("\n");
      if (ilp->index == 0) 
      {
        do_dump_prompt();
        return(FALSE);
      }
      ilp->buf[ilp->index++] = '\0';
      find_command(ilp);
      ilp->index = 0;
      return(TRUE);
    
    default:
      if (ilp->index >= LINE_LEN - 1)
        return(FALSE);
/*    printf("%c", rchar);   */
      ilp->buf[ilp->index++] = rchar;
      return(FALSE);
  }
}

/*****************************************************************************
 *                                                                           *
 * exec_cmd is called from within a dbd command to cause the                 *
 * execution of another command.  The first arg is the command               *
 * text.  The second arg gives the state of the looping flag                 *
 * with which the command should be run                                      *
 *                                                                           *
 *****************************************************************************/
exec_cmd(str, flag)
char *str;
int flag;
{
  struct input_line input;
  int ret;
  
  scopy(str, input.buf);
  find_command(&input);
  
  if (input.cm_type == UNDEFINED || input.cm_type == NULL_DUMP_CMD) 
  {
    printf("exec_cmd: unknown command: %s\n", str);
    return(FALSE);
  }
  
  input.loop = flag;
  
  ret = (*dcmd_def[input.cm_type].cmd_func)(&input);
  return(ret);
}



/****************************************************************************
 * find_command                                                             *
 ****************************************************************************/
find_command(ilp)
struct input_line *ilp;
{
  short i;
  register char *p, *q;
  extern char *find_arg(); 
  
  ilp->argc = 1;      /* the command itself counts as one arg */
  /*
   * any command can be prefixed with an exclamation
   * point to make it repeat until stopped.
   * Using two exclamation points makes it try harder to keep running,
   * ignoring more error conditions.
   */
  p = ilp->buf;
  ilp->non_stop = FALSE;
  if (*p == '!') 
  {
    ilp->loop = FIRST_TIME;
    p++;
    if (*p == '!') {
      ilp->non_stop = TRUE;
      p++;
    }
  } 
  else
    ilp->loop = DONT_LOOP;
  
  
  ilp->arg[0] = p;
  if (*p == '\0')
  {
    ilp->cm_type = NULL_DUMP_CMD;
    return(TRUE);
  }
  
  /*skip over verb to first arg */
  while(*p != ' '  &&  *p != '\0')
    p++;
  
  for (i = 1; i < N_ARGS; i++)
    p = find_arg(p, ilp, i);
  
  ilp->cm_type = search_dcmd(ilp->arg[0]);
  
  /*commands which have no arguments but do have a help message
   *clearly need help.
   */
  if (ilp->cm_type != UNDEFINED)
  {
    /* point to the full cmd name */
    ilp->arg[0] = dcmd_def[ilp->cm_type].cmd_name;
    
    if ((*ilp->arg[0] == '\0')  &&  (dcmd_def[ilp->cm_type].help_msg))
    {
      usage(ilp);
      ilp->cm_type = NULL_DUMP_CMD;
    }
  }
  
  return(TRUE);
}




/****************************************************************************
 * find_arg                                                                 *
 ****************************************************************************/
char *find_arg(p, ilp, index)
char *p;
struct input_line *ilp;
short index;
{
  char *q;
  
  /*skip over any white space*/
  while(*p == ' ')
    *p++ = '\0';
  
  ilp->arg[index] = p;
  q = p;
  
  /* if this is a real argument, bump up the count */
  if (*q)
    ilp->argc++;
  
  /* skip to end of current argument  */
  while(*p != ' '  &&  *p != '\0')
    p++;
  
  if (*p != '\0')
    *p++ = '\0';
  
  /* convert the arg to a number if possible (both decimal and hex) */
  if (*q) 
  {
    atoiv(q, &(ilp->decval[index]), 10);
    atoiv(q, &(ilp->val[index]), 16);
  } 
  else
    ilp->val[index] = 0;
  
  return(p);
}




/****************************************************************************
 * usage -- print a usage message for the given command                     *
 ****************************************************************************/
usage(ilp)
struct input_line *ilp;
{
  if (dcmd_def[ilp->cm_type].help_msg != NULL)
      printf("%s: %s\n", dcmd_def[ilp->cm_type].cmd_name,
              dcmd_def[ilp->cm_type].help_msg);
}




/****************************************************************************
 * search_dcmd                                                              *
 ****************************************************************************/
search_dcmd(test)
register char *test;
{
  register short matches, list_id;
  register short i;
  
  matches = 0;
  for (i = 0; dcmd_def[i].cmd_name != 0; i++)
    switch (smatch(test, dcmd_def[i].cmd_name)) 
    {
      case SUBSTRING:
        matches ++;
        list_id = i;
        break;
      
      case EXACT_MATCH:
        return(i);
    }
  
  if (matches == 1)
    return(list_id);
  return(UNDEFINED);
}




/****************************************************************************
 * console_cmd_init                                                         *
 ****************************************************************************/
console_cmd_init()
{
  printf("\n***   This is the Stress Test Command Processor   ***\n");
  printf("version %s, compiled %s.  \n", gracie_version, gracie_date);
  printf("Type ? for help\n");

  input.index = 0;

  do_dump_prompt();
}



/****************************************************************************
 * do_dump_prompt                                                           *
 ****************************************************************************/
do_dump_prompt()
{
  printf("Stress_Test[GP1000]-> ");
}




/****************************************************************************
 * character printing routine                                               *
 * calls device-dependent printer cpu_putc                                  *
 ****************************************************************************/
putc(c)
char c;
{
  int i;

  if (c == '\n') 
  {
    cpu_putc('\r');
    cpu_putc('\n');
#ifdef NEWLINE_DELAY
    for (i = 0; i < 8; i++)   /*pad the newline*/
      cpu_putc('\0');
#endif
  }
  else
    cpu_putc(c);
}


/****************************************************************************
 ** getc is a blocking read from the ascii console terminal                 *
 ****************************************************************************/
getc()
{
  int rchar;

  while ((rchar = cpu_getc()) == UNDEFINED);
  rchar &= 0x7f;
  return(rchar);
}


/***************************************************************************
 * application_init()                                                      *
 ***************************************************************************/
application_init()
{
  register int i, j;

/*
 * pid_array holds the names and pid's of children spawned by the
 * stress test during the course of testing.  Used by the 'kill'
 * command to get rid of those that are running in a forever loop.
 */

  for (i = 0; i < MAXNUMCHLD; i++)
  {
    pid_array[i].pid = 0;
    for (j = 0; j < MAXNAMLEN; j++)
      pid_array[i].name[j] = '\0';
  }
}
